exception Fatal of string
exception SCT_NOTFOUND

type argid = Syntax.vid
type subj = Syntax.vid
type posid = string
type lvid = int
type level = LVar of int (*** | Const of int ***)
let app_lsubst s x = Utilities.subst_var s x x
let comp_lsubst = Utilities.comp_subst app_lsubst

exception VerificationFailure of string * (lvid * lvid) list * (lvid, lvid) Utilities.substitution
type svid = argid * posid
type size_struct = S_NONE | S_CONST of int | S_EQ of svid | S_LT of svid | S_LEQ of svid
                   | S_PAIR of size_struct * size_struct

let rec non_decreasing s =
  match s with
    S_NONE -> true
  | S_CONST _ -> true
  | S_EQ _ -> true
  | S_LT _ -> false
  | S_LEQ _ -> false
  | S_PAIR(s1,s2) -> non_decreasing(s1) && non_decreasing(s2)

type exlevel = level * subj * size_struct (*** extended level information ***)

type single_rel = (level * subj * argid) * exlevel 

type mset_rel = (level * subj * argid) list * exlevel list 

type type4constr = mset_rel
type type3constr = mset_rel * type4constr option
type type2constr = mset_rel * type3constr option

type ordconstr = (subj * lvid) list * (subj * lvid) list

(*** transform type-4 constraints into order constraints ***)
let type4constr_to_ordc c =
  let f (mi, mo) = (List.map (fun (LVar l,x,_)->(x,l)) mi, List.map (fun (LVar l,x,_)->(x,l)) mo) in
     List.map f c

exception OrdConstr of ordconstr list * (lvid * lvid) list * (lvid, lvid) Utilities.substitution

exception InconsistentType2c of mset_rel
exception InconsistentType3c of mset_rel
exception UnsolvedType4c of type4constr list

let msetr2string (mi,mo) = 
       let inputs = List.map (fun (_,subj,_) -> subj) mi in
       let outputs = List.map (fun (_,subj,_) -> subj) mo in
       let seq_in = List.fold_left (fun s -> fun subj -> ((subj)^" "^s)) "" inputs in
       let seq_out = List.fold_left (fun s -> fun subj -> ((subj)^" "^s)) "" outputs in
            (seq_in^" > "^seq_out)

(*** init_constraints: mset_rel -> type2constr list ***)
let rec init_constraints constr: type2constr list =
  List.map (fun mset_r ->
                let (ms1,ms2) = mset_r in
                let n1 = List.length ms1 in
                  if n1 = List.length ms2 then (mset_r, Some(mset_r, Some mset_r))
                  else if n1=1 then (mset_r, None)
                  else (mset_r, Some(mset_r, None)))
            constr;;

let gen_sctfilename(scc) =
  "tmp"^string_of_int(List.hd(scc))^".sct";;
let gen_sctoutfilename(scc) =
  "tmp"^string_of_int(List.hd(scc))^".out";;

let check_termination (scc, ledges) =
  let get_argid x = x 
  in
  let rec gen_sctinput_sub l1 l2 s argid posid =
    match s with
      S_NONE -> ""
    | S_CONST n -> ""
    | S_EQ (argid',posid') -> if argid=argid' then "("^posid'^" "^posid^" deq) " else ""
    | S_LEQ (argid',posid') -> if argid=argid' then "("^posid'^" "^posid^" deq) " else ""
    | S_LT (argid',posid') -> if argid=argid' then "("^posid'^" "^posid^" dec) " else ""
    | S_PAIR(s1,s2) -> ((gen_sctinput_sub l1 l2 s1 argid ("fst_"^posid))^" "
                         ^(gen_sctinput_sub l1 l2 s2 argid ("snd_"^posid)))
  in
  let gen_sctinput l1 l2 s1 s2 =
    let argid = get_argid s1 in
       ("(level"^(string_of_int l1)^" level"^(string_of_int l2)^" {"^
         (gen_sctinput_sub l1 l2 s2 argid "x")^"})")
  in
  let gen_sctsub ((LVar l1, _, s1), (LVar l2,_, s2)) = 
        (gen_sctinput l1 l2 s1 s2)
  in
  let sct_trs = List.fold_left (fun sctin -> fun tr -> (gen_sctsub tr)^sctin) "" ledges in
  let sctinput = "{\n"^sct_trs^"\n}\n" in
  let sctfile = gen_sctfilename(scc) in
  let oc = open_out sctfile in
  let _ = output_string oc sctinput in
  let _ = close_out oc in
  let sctoutfile = gen_sctoutfilename(scc) in
  let _ = if Sys.command ("sct <"^sctfile^" >"^sctoutfile) =0 
          then () 
          else raise SCT_NOTFOUND
  in
  let ic = open_in sctoutfile in
  let result = input_line ic in
  let _ = close_in ic in
    result="#T"

(*** {a[s1]} \cup S1 > S2 and a is in a cycle -> {a}\cup S1 > S2 ***)
(*** {a}\cup S1 > {a}\cup S2 --> S1 > S2 ***)
let merge_levels_in_constr cycles type2constrs type4constrs =
  let rec unify_levels_in_elevel_left cycles (LVar(l1),subj,s) = 
    match cycles with
       [] -> (LVar(l1),subj,s)
     | cyc::cycles' -> 
         if List.mem l1 cyc then (LVar(List.hd(cyc)), subj, s)
         else unify_levels_in_elevel_left cycles' (LVar(l1),subj, s)
  in
  let rec unify_levels_in_elevel_right cycles (LVar(l1),subj,s) = 
    match cycles with
       [] -> (LVar(l1),subj, s)
     | cyc::cycles' -> 
         if List.mem l1 cyc then (LVar(List.hd(cyc)), subj, S_NONE)
         else unify_levels_in_elevel_right cycles' (LVar(l1),subj, s)
  in
  let unify_levels_in_type4c cycles (ms1,ms2) = 
    let ms1' = List.map (unify_levels_in_elevel_left cycles) ms1 in
    let ms2' = List.map (unify_levels_in_elevel_right cycles) ms2 in
      (ms1',ms2')
  in
  let unify_levels_in_type3c cycles ((ms1,ms2),c) = 
    let ms1' = List.map (unify_levels_in_elevel_left cycles) ms1 in
    let ms2' = List.map (unify_levels_in_elevel_right cycles) ms2 in
    match c with
      Some c1 -> ((ms1',ms2'),Some (unify_levels_in_type4c cycles c1))
    | None -> ((ms1',ms2'),None)
  in
  let unify_levels_in_type2c cycles (((ms1,ms2),c): type2constr) = 
    let ms1' = List.map (unify_levels_in_elevel_left cycles) ms1 in
    let ms2' = List.map (unify_levels_in_elevel_right cycles) ms2 in
    match c with
      Some c1 -> ((ms1',ms2'), Some (unify_levels_in_type3c cycles c1))
    | None -> ((ms1',ms2'),None)
  in
  let occurs_in l ms =
    List.exists (fun (LVar(l1),_,s1) -> l1=l && non_decreasing(s1)) ms
  in
  let rec delete l ms =
    match ms with
      [] -> raise Not_found
    | (LVar(l1),subj1,s1)::ms' -> 
          if l=l1 && non_decreasing(s1) then ms'
          else (LVar(l1),subj1,s1)::(delete l ms')
  in
  let rec is_cyclic ms1 ms2 =
    match ms1 with
       [] -> false
     | (LVar(l1),_,_)::ms1' -> (occurs_in l1 ms2)||(is_cyclic ms1' ms2)
  in
  let rec diff ms1 ms2 =
    match ms1 with
       [] -> ([], ms2)
     | (LVar(l1),subj1,s1)::ms1' ->
         try 
           let ms2' = delete l1 ms2 in
             diff ms1' ms2'
         with
           Not_found -> let (ms1'', ms2'') = diff ms1' ms2 in ((LVar(l1),subj1,s1)::ms1'', ms2'')
  in
  let simplify_type3c (((ms1,ms2),c):type3constr) =
    let (ms1',ms2') = diff ms1 ms2 in
      if ms1'=[] then
         if ms2'!=[] then
              raise (InconsistentType3c (ms1,ms2))
         else 
           match c with
              None -> raise (Fatal "simplify_type3c")
            | Some c1 -> ([], [c1])
      else if ms2' = [] then
           ([], [])   (*** type 3 constraint has been solved ***)
      else
         ([((ms1',ms2'), Some((ms1',ms2'), c))], []) (*** new type 2 constraint ***)
  in
  let merge_levels_in_type2c cycles (type2constr:type2constr) = 
    let ((ms1,ms2),c) = unify_levels_in_type2c cycles type2constr in
      if is_cyclic ms1 ms2 then
         match c with
           Some c1 -> simplify_type3c c1
         | None -> raise (InconsistentType3c (ms1,ms2))
      else ([((ms1,ms2),c)], [])
  in
  let (new_type2cs, new_type4cs) =
           List.fold_left
             (fun (type2cs, type4cs) -> fun type2c ->
                 let (x,y) = merge_levels_in_type2c cycles type2c in
                    (x@type2cs, y@type4cs)) ([], []) type2constrs in
  let new_type4cs2 = List.map (unify_levels_in_type4c cycles) type4constrs in
      (new_type2cs, new_type4cs2@new_type4cs)

let print_constraints type2cs =
  let print_ms ms =
     let _ = List.map (fun (LVar(l),_,_) -> (print_int l; print_string " ")) ms in ()
  in
  let print_type2c ((ms1,ms2),c) =
       (print_ms ms1; print_string " > "; print_ms ms2; print_string "\n")
  in
  let _ = List.map print_type2c type2cs in ()

(*** {a,b} > S /\ a<b ==> b>S ***)
let reduce_type2constrs (rep_map: (lvid * lvid) list) rgraph (type2cs: type2constr list) =
  let filter_out_reachables l ms =
    let rset = try Scc.compute_rset rgraph (List.assoc l rep_map) with
                 Not_found -> [l] in
     (List.filter (fun (LVar x,_,_) -> 
                    let x' = try List.assoc x rep_map with Not_found -> x
                    in not(List.mem x' rset)) ms)
  in
  let rec reduce_ms ms checked=
    match ms with
      [] -> checked
    | (LVar l1,subj,s)::ms' ->
         let ms'' = filter_out_reachables l1 ms' in
         let checked' = filter_out_reachables l1 checked in
            reduce_ms ms'' ((LVar l1,subj,s)::checked')
  in
  let rec reduce_sub (type2cs:type2constr list) newcs flag =
     match type2cs with
        [] -> (newcs, flag)
      | ((ms1,ms2),c)::type2cs' ->
           match ms1 with
             [_] -> reduce_sub type2cs' (((ms1,ms2),c)::newcs) flag
           | _ ->
               let ms1' = reduce_ms ms1 [] in
               let flag' = flag || (List.length ms1'=1) in
                  reduce_sub type2cs' (((ms1',ms2),c)::newcs) flag'
  in
     reduce_sub type2cs [] false

let type2cs_to_srels type2cs = 
       List.fold_left
                    (fun rels -> fun ((ms1,ms2),_) ->
                      match ms1 with
                       [(l1,subj1,s1)] -> 
                           List.fold_left 
                             (fun r->fun (l2,subj2,s2)->((l1,subj1,s1),(l2,subj2,s2))::r) rels ms2 
                      | _ -> rels)
                    [] type2cs

let srels_to_edges srels = 
  Utilities.delete_duplication_unsorted 
               (List.map (fun ((LVar l1,_,_),(LVar l2,_,_)) -> (l1,l2)) srels)

let type2cs_to_edges type2cs =
    srels_to_edges(type2cs_to_srels type2cs)

let rec reduce_type2constraints (type2constr: type2constr list) (type4constr: type4constr list) lsubst =
  let single_rels = type2cs_to_srels type2constr in 
  let edges = srels_to_edges single_rels in 
  let sccs = Scc.compute_scc edges in
  let sccs_with_ledges =
     List.map
     (fun scc ->
        (scc, List.filter (fun ((LVar l1,_,_),(LVar l2,_,_))-> (List.mem l1 scc && List.mem l2 scc)) single_rels))
     sccs in
  let sccs_with_ledges' =
     List.filter (fun (scc, rels) -> rels!=[]) sccs_with_ledges in
  let sct_result = List.combine sccs_with_ledges' (List.map check_termination sccs_with_ledges') in 
  let cycles = List.map (fun x -> fst(fst x)) (List.filter (fun (scc,r)->r=false) sct_result) in
  let lsubst1 = List.flatten 
                 (List.map 
                   (fun cycle -> let rep = List.hd cycle in List.map (fun y->(y,rep)) 
                   (List.tl cycle))
                    cycles) in
  let lsubst2 = comp_lsubst lsubst1 lsubst in
  let (type2constr', type4constr') =
           try
              merge_levels_in_constr cycles type2constr type4constr 
           with
             InconsistentType3c (mi, mo) ->
                let message = "Unsolvable multiset constraints:\n   "^(msetr2string(mi, mo)) in
                   raise (VerificationFailure (message, edges, lsubst2))
  in
  let rep_map = List.flatten (List.map (fun scc -> List.map (fun x -> (x,List.hd(scc))) scc) sccs) in
  let edges_between_sccs = List.map (fun (l1,l2)->(List.assoc l1 rep_map, List.assoc l2 rep_map))
                           edges in
  let acyclic_edges = Utilities.delete_duplication_unsorted
                       (List.filter (fun (l1,l2)->l1!=l2) edges_between_sccs) in
  let rgraph = Scc.mk_rgraph acyclic_edges in
  let (type2cs'', flag) = reduce_type2constrs rep_map rgraph type2constr' in
    (print_string "one step completed\n";
     print_constraints type2cs'';
    if (flag || (cycles!=[])) then 
       (reduce_type2constraints type2cs'' type4constr' lsubst2)
    else
       let 
         (solved, unsolved) = List.partition (fun ((mi,_),_) -> List.length mi = 1) type2cs'' 
       in
         if unsolved = [] 
         then if type4constr' = [] 
              then
                 (edges, lsubst2)
              else
                let ordcs = type4constr_to_ordc type4constr' in
(***
 ***            let m = List.fold_left (fun s -> fun type4c -> msetr2string(type4c)^"\n   "^s) "" 
 ***                     type4constr' in
 ***            let message = 
 ***               "Unsolvable multiset constraints (May be solved with 4th type system) :\n   "^m in
 ***)
                raise (OrdConstr (ordcs, edges, lsubst2))
         else
           let ((mi,mo),type3c_opt) = List.hd unsolved in
           let rest = List.tl unsolved in
           let new_type2c = (([List.hd(mi)],mo), type3c_opt) in 
                         (*** pick the first input as a candidate of the highest leve input ***) 
           let new_type2cs = new_type2c::(rest@solved) in
             reduce_type2constraints new_type2cs type4constr' lsubst2);;


let solve_constraints c =
 try
   let type2c = init_constraints c in
   let (type2c',type4c) = 
          try
             merge_levels_in_constr [] type2c [] 
          with
                InconsistentType3c (mi, mo) ->
                let message = "Unsolvable multiset constraints:\n   "^(msetr2string(mi, mo)) in
                let edges = type2cs_to_edges type2c in
                   raise (VerificationFailure (message, edges, Utilities.subst_empty))
   in
   let (edges, lsubst) = reduce_type2constraints type2c' type4c [] in
      (edges, lsubst, None)
 with
   VerificationFailure (s, edges, lsubst) ->
      (edges, lsubst, Some(s))

(*** sample inputs ***)

let scc1 = [1;2;3]

let edges = [((1, "x"), (2, S_PAIR (S_LT ("y", "fst_x"), S_LT("x", "fst_x"))));
             ((2, "y"), (3, S_LT ("y", "snd_x")));
             ((3, "y"), (1, S_PAIR (S_LEQ("y", "x"), S_NONE)))];;

let cs1 = [([(1, "x")], [(2, S_NONE)])];;

let cs2 = [([(1, "x"); (2, "y")], [(1,S_NONE);(1,S_NONE); (3,S_NONE)]);
           ([(1, "x"); (2, "y")], [(1, S_NONE); (4, S_NONE)]);
           ([(4, "z")], [(2, S_NONE)])];;
