open Syntax;;

type rectyId = int
type tvid = int
type nsvid = int
type nsubst = (string * string list) list 
        (*** A name substitution is a mapping from strings to sets of string, rather than to strings.
         *** That is for handling lists, trees, etc.
         ***)
type names = NVar of nsvid | NSubst of nsubst * names | NConst of string list
type otype = UnitT
           | BoolT
           | IntT
           | ChanT of otype * names
           | PairT of otype * otype
           | FunT of otype * otype
           | RecT of rectyId * otype list 
type tenv = (Syntax.vid, otype) Utilities.env

type name_struct = NS_empty | NS_name of string | NS_pair of name_struct * name_struct
                 | NS_set of name_struct list
type venv = (Syntax.vid, name_struct) Utilities.env

exception TODO;;
exception Error of string;;
exception Fatal of string;;

type constr = EqNames of nsvid * nsvid
            | SubNS of nsvid * nsubst * nsvid   
               (*** SubNS(x, s, y) <==> x \supseteq s y ***)
            | SubNS2 of nsubst * nsvid * nsvid
            | UpBound of nsvid * string list

type constraints = constr list

let nset_union = Utilities.merge_and_unify compare;;

let rec nset_of_ns ns =
  match ns with
    NS_empty -> []
  | NS_name x -> [x]
  | NS_pair(ns1,ns2) -> nset_union (nset_of_ns ns1) (nset_of_ns ns2)
  | NS_set nss ->
      List.fold_left (fun nset -> fun ns1 -> nset_union nset (nset_of_ns ns1)) [] nss

let nset_of_venv venv =
  let ns_set = List.map (fun (x,ns) -> ns) venv in
    nset_of_ns (NS_set ns_set)

let union_nvars = Utilities.merge_and_unify compare;;

let nvars_in_ac ac =
  match ac with
    EqNames(v1,v2) -> union_nvars [v1] [v2]
  | SubNS(v1,_,v2) -> union_nvars [v1] [v2]
  | SubNS2(_,v1,v2) -> union_nvars [v1] [v2]
  | UpBound(v,_) -> [v]

let rec nvars_in_c c =
  match c with
    [] -> []
  | ac::c' -> union_nvars (nvars_in_ac ac) (nvars_in_c c')

let merge_constraints = List.fold_left (@) []

let emptyC = []

let current_nsvid = ref 0;;
let new_nsvid() = 
    let x = !current_nsvid in
    let _ = (current_nsvid := !current_nsvid+1) in
       x;;

let alias_tab = ref []

let register_alias x x' =
   alias_tab := (x,x')::!alias_tab

let reset() = (current_nsvid := 0; alias_tab := [])

let dummy_pc = -1;;

let rec new_otype t =
  match t with
    Stype.TVar n -> UnitT
  | Stype.UnitT -> UnitT
  | Stype.BoolT -> BoolT
  | Stype.IntT -> IntT
  | Stype.ChanT t1 -> let l1 = new_nsvid() in
                      (ChanT(new_otype t1, NVar l1))
  | Stype.PairT(t1,t2) -> PairT(new_otype t1, new_otype t2)
  | Stype.FunT(t1,t2) -> FunT(new_otype t1, new_otype t2)
  | Stype.RecT(id,tl) -> RecT(id, List.map new_otype tl)

let const_table = [("+", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("-", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("*", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("/", fun _ -> FunT(PairT(IntT, IntT),IntT));
                   ("=", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   (">", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   (">=", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   ("<", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   ("<=", fun _ -> FunT(PairT(IntT, IntT),BoolT));
                   ("fst", fun t -> 
                            match t with
                               Stype.FunT(Stype.PairT(t1,t2),t3) -> 
                                 let t1' = new_otype t1 in
                                 let t2' = new_otype t2 in
                                     FunT(PairT(t1',t2'),t1')
                             | _ -> raise (Fatal "const_table"));
                   ("snd", fun t -> 
                            match t with
                               Stype.FunT(Stype.PairT(t1,t2),t3) -> 
                                 let t1' = new_otype t1 in
                                 let t2' = new_otype t2 in
                                     FunT(PairT(t1',t2'),t2')
                             | _ -> raise (Fatal "const_table"));
                   ("nil", fun t -> new_otype t);
                   ("cons", fun t ->
                             match t with
                               Stype.FunT(Stype.PairT(t1,Stype.RecT(_, [t2;_])),_) ->
                                    let t1' = new_otype t1 in
                                    let t2' = new_otype t2 in
                                        (FunT(PairT(t1', RecT(0, [t2';t2'])), RecT(0, [t1';t2'])))
                             | _ -> raise (Fatal "const_table:cons"));
                   ("hd", fun t -> 
                           match t with
                               Stype.FunT(Stype.RecT(_,[t1;t2]), _) ->
                                    let t1' = new_otype t1 in
                                    let t2' = new_otype t2 in
                                       FunT(RecT(0,[t1';t2']), t1')
                             | _ -> raise (Fatal "const_table:hd"));
                   ("tl", fun t -> 
                           match t with
                               Stype.FunT(Stype.RecT(_,[t1;t2]), _) ->
                                    let t1' = new_otype t1 in
                                    let t2' = new_otype t2 in
                                       FunT(RecT(0,[t1';t2']), RecT(0,[t2';t2']))
                             | _ -> raise (Fatal "const_table:hd"));
                   ("null", fun t -> 
                           match t with
                               Stype.FunT(Stype.RecT(_,[t1;t2]),_) ->
                                  let t1' = new_otype t1 in
                                  let t2' = new_otype t2 in
                                      FunT(RecT(0, [t1';t2']), BoolT)
                            | _ -> raise (Fatal "const_table:null"))];;

let lookup_prim prname t = 
  try
    (List.assoc prname const_table) t
  with
    Not_found -> raise (Fatal ("unknown primitive "^prname));;

let prim_nsfun_tab =
                  [("+", fun _ -> NS_empty);
                   ("-", fun _ -> NS_empty);
                   ("*", fun _ -> NS_empty);
                   ("/", fun _ -> NS_empty);
                   ("=", fun _ -> NS_empty);
                   (">", fun _ -> NS_empty);
                   (">=", fun _ -> NS_empty);
                   ("<", fun _ -> NS_empty);
                   ("<=", fun _ -> NS_empty);
                   ("fst", function NS_pair(ns1,ns2) -> ns1 | _ -> raise (Fatal "prim_nsfun_tab"));
                   ("snd", function NS_pair(ns1,ns2) -> ns2 | _ -> raise (Fatal "prim_nsfun_tab"));
                   ("nil", fun _ -> NS_empty);
                   ("cons", fun ns ->
                             match ns with
                               NS_pair(ns1,ns2) -> 
                                  (match ns2 with
                                    NS_empty -> NS_set [ns1] 
                                  | NS_set(nss) -> NS_set (ns1::nss)
                                  | _ -> raise (Fatal "prim_nsfun"))
                             | _ -> raise (Fatal "prim_nsfun"));
                   ("hd", fun ns ->  
                           match ns with
                              NS_empty -> NS_empty
                             | NS_set(nss) -> List.hd(nss) (*** This part should be revised? ***)
                             | _ -> raise (Fatal "prim_nsfun:hd"));
                   ("tl", fun ns -> 
                           match ns with
                              NS_empty -> NS_empty
                             | NS_set([ns1]) -> NS_set([ns1])(*** This part should be revised? ***)
                             | NS_set(nss) -> NS_set(List.tl(nss))
                             | _ -> raise (Fatal "prim_nsfun:tl"));
                   ("null", fun _ -> NS_empty)];;
    
let lookup_prim_nsfun s = List.assoc s prim_nsfun_tab;;

let string_of_names names =
  match names with
    NVar n -> "S"^(string_of_int n)
  | NConst nl  -> "{"^(List.fold_left (fun s -> fun n -> n^" "^s) "" nl)^"}"
  | NSubst _ -> "..."

let rectype_print_table =
  ref [(0, fun f -> fun tl -> ((f 3 (List.hd(tl)))^" list"))]

let string_of_type t =
  let rec t2string pr t = 
    match t with
      UnitT -> "unit"
    | BoolT -> "bool"
    | IntT -> "nat"
    | ChanT(t1, names) -> "chan("^(t2string 0 t1)^", "^(string_of_names names)^")"
    | PairT(t1,t2) -> 
           (if pr>1 then "(" else "")
           ^(t2string 2 t1)^"*"^(t2string 2 t2)
          ^(if pr>2 then ")" else "")
    | FunT(t1,t2) -> 
           (if pr>0 then "(" else "")
          ^(t2string 1 t1)
          ^"->"
          ^(t2string pr t2)
          ^(if pr>0 then ")" else "")
    | RecT(id, tl) ->
        let f = List.assoc id !rectype_print_table in
          f t2string tl
  in t2string 0 t;;

let print_env env = Utilities.print_env (fun x->x) string_of_type env;;

let tenv_extend x tx env =
    Utilities.env_extend x tx env;;
let venv_extend x ns venv =
    Utilities.env_extend x ns venv;;
let venv_lookup = Utilities.env_lookup

exception InvalidMerge of nsubst list
let rec merge_name_substitution ss =
  match ss with
    [] -> []
  | [s] -> s
  | s::ss' -> 
       let s2 = merge_name_substitution ss' in
       let x = List.combine s s2 in
         List.map (fun ((n1,ns1), (n2,ns2)) -> 
                   if n1=n2 then (n1, ns1@ns2) 
                   else raise (InvalidMerge (ss))) x

exception InvalidNsubst of name_struct * name_struct
let rec mk_nsubst (ns1,ns2) = 
  match (ns1,ns2) with
    (NS_empty, _) -> []
  | (NS_name v1, NS_name v2) -> [(v1, [v2])]
  | (NS_name v1, NS_empty) -> [(v1, [])]
  | (NS_pair(ns11,ns12),NS_pair(ns21,ns22)) -> (mk_nsubst(ns11,ns21))@(mk_nsubst(ns12,ns22))
  | (NS_pair(ns11,ns12),NS_empty) -> (mk_nsubst(ns11,NS_empty))@(mk_nsubst(ns12,NS_empty))
  | (NS_set([ns11]), NS_set(nss2)) -> 
           let ss = List.map (fun ns -> (mk_nsubst (ns11,ns))) nss2 in
             merge_name_substitution ss
  | (NS_set([ns11]), NS_empty) -> mk_nsubst(ns11, NS_empty)
  | _ -> raise (InvalidNsubst (ns1,ns2))
  
exception InvalidSubT of otype * otype

let rec makeEqT t1 t2 =
  match (t1,t2) with
    (UnitT,UnitT) -> []
  | (BoolT,BoolT) -> []
  | (IntT,IntT) -> []
  | (ChanT(t1',NVar l1), ChanT(t2',NVar l2)) ->
           (EqNames (l1, l2))::(makeEqT t1' t2')
  | (PairT(t11,t12), PairT(t21,t22)) ->
         (makeEqT t11 t21)@(makeEqT t12 t22)
  | (FunT(t11,t12), FunT(t21,t22)) ->
         (makeEqT t11 t21)@(makeEqT t12 t22)
  | (RecT(id1, tl1), RecT(id2,tl2)) ->
      if id1=id2 then
        let tpl = List.combine tl1 tl2 in
          List.fold_left (fun c -> fun (t1,t2) -> (makeEqT t1 t2)@c) [] tpl
      else raise (Fatal "makeEqT")
  | _ -> raise (Fatal "makeEqT")

let rec mkC_subT (t1,ns1,t2,ns2) = 
  match (t1,t2) with
    (UnitT,UnitT) -> []
  | (BoolT,BoolT) -> []
  | (IntT,IntT) -> []
  | (ChanT(t1',NVar l1), ChanT(t2',NVar l2)) ->
           (SubNS(l1, mk_nsubst(ns2,ns1), l2))::(makeEqT t1' t2')
  | (PairT(t11,t12), PairT(t21,t22)) ->
         (mkC_subT(t11,ns1,t21,ns2))@(mkC_subT(t12,ns1,t22,ns2))
  | (FunT(t11,t12), FunT(t21,t22)) ->
         (makeEqT t11 t21)@(makeEqT t12 t22)
  | (RecT(id1, tl1), RecT(id2,tl2)) ->
      if id1=id2 then
        let tpl = List.combine tl1 tl2 in
          List.fold_left (fun c -> fun (t1,t2) -> (mkC_subT(t1,ns1,t2,ns2))@c) [] tpl
      else raise (InvalidSubT (t1, t2))
  | _ -> raise (InvalidSubT (t1, t2))
     
let rec mkC_subT2 (t1,ns1,t2,ns2) = 
  match (t1,t2) with
    (UnitT,UnitT) -> []
  | (BoolT,BoolT) -> []
  | (IntT,IntT) -> []
  | (ChanT(t1',NVar l1), ChanT(t2',NVar l2)) ->
           (SubNS2(mk_nsubst(ns1,ns2), l1, l2))::(makeEqT t1' t2')
  | (PairT(t11,t12), PairT(t21,t22)) ->
         (mkC_subT2(t11,ns1,t21,ns2))@(mkC_subT2(t12,ns1,t22,ns2))
  | (FunT(t11,t12), FunT(t21,t22)) ->
         (makeEqT t11 t21)@(makeEqT t12 t22)
  | (RecT(id1, tl1), RecT(id2,tl2)) ->
      if id1=id2 then
        let tpl = List.combine tl1 tl2 in
          List.fold_left (fun c -> fun (t1,t2) -> (mkC_subT2(t1,ns1,t2,ns2))@c) [] tpl
      else raise (Fatal "makeSubT2")
  | _ -> raise (Fatal "makeSubT2")
let rec mkNS x t =
  match t with
    UnitT -> NS_empty
  | BoolT -> NS_empty
  | IntT -> NS_empty
  | ChanT(_) -> NS_name(x)
  | PairT(t1,t2) -> 
      NS_pair(mkNS ("fst#"^x) t1, mkNS ("snd#"^x) t2)
  | FunT(t1,t2) -> NS_empty
  | RecT(id,tl) -> mkNS_rect id x tl
and mkNS_rect id x tl =
  match id with 
    0 -> let t = List.hd(tl) in let ns1 = mkNS ("elem#"^x) t in NS_set([ns1])
  | _ -> raise (Fatal "mkNS_rect")

let t2ns t = mkNS "." t

let mkC_NSbound(nsvid, nset) = [UpBound(nsvid, nset)]
let top_names t = nset_of_ns(t2ns t)
let rec mkC_Tbound(t, nset_g, nset_l, ownname) =
  match t with
     UnitT -> []
   | BoolT -> []
   | IntT -> []
   | ChanT(t1, NVar id) -> 
         let nset_l' = Utilities.list_diff nset_l [ownname] in
         let c1 = mkC_NSbound(id, nset_g@nset_l') in
         let nset_l'' = top_names t1 in
         let c2 = mkC_Tbound(t1, nset_g, nset_l'', "@") in
          merge_constraints [c1;c2]
   | PairT(t1,t2) -> 
        merge_constraints [mkC_Tbound(t1,nset_g, nset_l, "fst_"^ownname); 
                           mkC_Tbound(t2,nset_g, nset_l, "snd_"^ownname)]
   | FunT(t1,t2) -> 
        merge_constraints [mkC_Tbound(t1,nset_g, [], "@"); 
                           mkC_Tbound(t2,nset_g, [], "@")]
   | RecT(id, tl) ->
       if id=0 then
          merge_constraints (List.map (fun t -> mkC_Tbound(t,nset_g, nset_l, "elem_"^ownname)) tl)
       else raise (Fatal "mkC_Tbound")
   | _ -> raise (Fatal "mkC_Tbound")

let rec mkC_Tboundempty(t) =
  match t with
     UnitT -> []
   | BoolT -> []
   | IntT -> []
   | ChanT(t1, NVar id) -> 
         let c1 = mkC_NSbound(id, []) in
         let c2 = mkC_Tboundempty(t1) in
          merge_constraints [c1;c2]
   | PairT(t1,t2) -> merge_constraints [mkC_Tboundempty(t1); mkC_Tboundempty(t2)]
   | FunT(t1,t2) -> merge_constraints [mkC_Tboundempty(t1); mkC_Tboundempty(t2)]
   | RecT(id, tl) ->
       merge_constraints (List.map (fun t -> mkC_Tboundempty(t)) tl)
   | _ -> raise (Fatal "mkC_Tboundempty")



(*** tinf: stype procexp -> tenv -> venv -> otype procexp * constraints ***)
let rec tinf (p: Stype.stype procexp) env (venv:venv) nset =
  match p with
    Zero -> (Zero, emptyC)
  | Out(x,v,p1,pc) ->
      let (v', tv, ns1, c0) = tinfV v env venv in
      let (p1',c1) = tinf p1 env venv nset in
      let tx = tinfVar x env in
      let tx_arg = match tx with
                 ChanT(t,_) -> t
               | _ -> raise (Fatal "tinf") in
      let ns2 = t2ns tx_arg in
      let c2 = mkC_subT(tv, ns1, tx_arg, ns2) in  (*** tv < [ns1/ns2]tx_arg ***)
      let c = merge_constraints [c0;c1;c2] in
        (Out(x,v',p1',pc), c)
  | In(x,y,t1,p1,pc) ->
      let tx = tinfVar x env in
      let tx_arg = match tx with 
                       ChanT(t,_) -> t
                     | _ -> raise (Fatal "tinf") in
      let ns1 = t2ns tx_arg in
      let ty = new_otype t1 in
      let env1 = tenv_extend y ty env in
      let ns2 = mkNS y ty in
      let venv1 = venv_extend y ns2 venv in
      let nset1 = nset_union nset (nset_of_ns ns2) in
      let (p1',c1) = tinf p1 env1 venv1 nset1 in
      let c2 = mkC_subT2(tx_arg, ns1, ty, ns2) in (*** [ns2/ns1]tx_arg < ty ***)
      let c3 = mkC_Tbound(ty, nset1, [], "@") in
      let c = merge_constraints [c1;c2;c3] in
        (In(x,y,ty,p1',pc),c)
  | Par(p1,p2) ->
      let (p1',c1) = tinf p1 env venv nset in
      let (p2',c2) = tinf p2 env venv nset in
      let c = merge_constraints [c1;c2] in
       (Par(p1',p2'), c)
  | Rep(p1) ->
      let (p1', c1) = tinf p1 env venv nset in
           (Rep(p1'), c1)
  | Nu(x,stx,p1,pc) ->
      let tx = new_otype stx in
      let nset1 = nset_union nset [x] in
      let env1 = tenv_extend x tx env in
      let venv1 = venv_extend x (NS_name(x)) venv in
      let (p1',c1) = tinf p1 env1 venv1 nset1 in
      let c2 = mkC_Tbound(tx, nset, [], x) in
      let c = merge_constraints [c1;c2] in
        (Nu(x,tx,p1',pc), c)
  | If(v, p1, p2, pc) ->
      let (v',tv,nsv,c0) = tinfV v env venv in
      let (p1',c1) = tinf p1 env venv nset in
      let (p2',c2) = tinf p2 env venv nset in
      let c = merge_constraints [c0;c1;c2] in
       (If(v',p1',p2',pc), c)
  | Let(x,t1,v,p1,pc) ->
      let (v',tv, nsv,c0) = tinfV v env venv in
      let env1 = tenv_extend x tv env in
      let venv1 = venv_extend x nsv venv in
      let _ = match nsv with
                NS_name x' -> register_alias x x'
              | _ -> () in
      let (p1',c1) = tinf p1 env1 venv1 nset in 
      let c = merge_constraints [c0;c1] in
        (Let(x,tv,v', p1',pc), c)
  | MatchL(x,matchlist,pc) -> 
      let tx = tinfVar x env in
      let (matchlist',c) = tinfMatchL tx matchlist env venv nset in
        (MatchL(x,matchlist',pc), c)
  | Special(p1) ->
      let (p1', c1) = tinf p1 env venv nset in
           (Special(p1'), c1)
and tinfMatchL t matchlist env venv nset =
   match matchlist with
      [] -> ([], emptyC)
    | (cid, p1)::matchlist1 ->
         let (matchlist1',c1) = tinfMatchL t matchlist1 env venv nset in
         let (p1',c2) = tinf p1 env venv nset in
         let c = merge_constraints [c1;c2] in
            ((cid,p1')::matchlist1', c)
and tinfVar x env =
   try 
     Utilities.env_lookup x env
   with
     Not_found -> raise (Fatal ("unknown variables "^x))
and tinfV v env venv = 
   match v with
     Var(x) -> (Var(x), tinfVar x env, venv_lookup x venv, emptyC)
   | Unit -> (Unit, UnitT, NS_empty, emptyC)
   | SBool(b,sl) -> (SBool(b,sl), BoolT, NS_empty,emptyC)
   | SInt(b,sl) -> (SInt(b,sl), IntT, NS_empty,emptyC)
   | Bool(b) -> (Bool(b), BoolT, NS_empty,emptyC)
   | Int(n) -> (Int n, IntT, NS_empty, emptyC)
   | Prim(s,ot) -> 
        (match ot with
           Some t1 -> let t = lookup_prim s t1 in (Prim(s,Some t), t, NS_empty, emptyC)
         | None -> raise (Fatal "tinfV")
         )
   | App(Prim(s,ot), v2, pc) ->
        let (v1',t1, _, c1) = tinfV (Prim(s,ot)) env venv in
        let (v2',t2, ns2, c2) = tinfV v2 env venv in
        let (t11,t12) = match t1 with
                         FunT(t11,t12) -> (t11,t12)
                         | _ -> raise (Fatal "tinfV") in
        let c3 = makeEqT t2 t11 in
        let ns = (lookup_prim_nsfun s) ns2 in
        let c = merge_constraints [c1;c2;c3] in
          (App(Prim(s,None), v2',pc), t12, ns, c)
   | App (_,_,_) -> raise (Fatal "tinfV")
   | Pair(v1,v2) -> 
        let (v1',t1, ns1, c1) = tinfV v1 env venv in
        let (v2',t2, ns2, c2) = tinfV v2 env venv in
        let c = merge_constraints [c1;c2] in
          (Pair(v1',v2'),PairT(t1,t2), NS_pair(ns1,ns2), c)


let app_nsubst s x = Utilities.subst_var s x x
let comp_nsubst = Utilities.comp_subst app_nsubst

let rec app_nsubst_for_t s t =
  match t with
    UnitT -> UnitT
  | BoolT -> BoolT
  | IntT -> IntT
  | PairT(t1,t2) -> PairT(app_nsubst_for_t s t1, app_nsubst_for_t s t2)
  | FunT(t1,t2) -> FunT(app_nsubst_for_t s t1, app_nsubst_for_t s t2)
  | RecT(id,tl) -> RecT(id, List.map (app_nsubst_for_t s) tl)
  | ChanT(t1,NVar vid) -> ChanT(app_nsubst_for_t s t1, NVar(app_nsubst s vid))
  | ChanT(t1, NConst x) -> ChanT(app_nsubst_for_t s t1, NConst x)
  | _ -> raise (Fatal "app_nsubst_for_t")

let rec subst_nconst_for_t s t =
  match t with
    UnitT -> UnitT
  | BoolT -> BoolT
  | IntT -> IntT
  | PairT(t1,t2) -> PairT(subst_nconst_for_t s t1, subst_nconst_for_t s t2)
  | FunT(t1,t2) -> FunT(subst_nconst_for_t s t1, subst_nconst_for_t s t2)
  | RecT(id,tl) -> RecT(id, List.map (subst_nconst_for_t s) tl)
  | ChanT(t1,NVar vid) -> ChanT(subst_nconst_for_t s t1, NConst (List.assoc vid s))
  | ChanT(t1, NConst x) -> ChanT(subst_nconst_for_t s t1, NConst x)
  | _ -> raise (Fatal "subst_nconst_for_t")

(*** reduce_eqL:
 ***  constr list ->
 ***   (int, int) Utilities.substitution *
 ***   (int * int * Syntax.vid * ttype Syntax.valexp *
 ***   (Syntax.vid * ttype Syntax.valexp) list * ttype Syntax.valexp list) list
 ***)

let reduce_eqN c =
  let app_subst_for_ac s ac =
    match ac with
      EqNames(n1, n2) -> EqNames(app_nsubst s n1, app_nsubst s n2)
    | SubNS(v1, nsubst, v2) ->
        SubNS(app_nsubst s v1, nsubst, app_nsubst s v2)
    | SubNS2(nsubst, v1, v2) ->
        SubNS2(nsubst, app_nsubst s v1, app_nsubst s v2)
    | UpBound(v, nsubst) ->
        UpBound(app_nsubst s v, nsubst)
  in
  let app_subst_for_c s c =
    List.map (app_subst_for_ac s) c 
  in
  let rec reduce_it c s reduced =
    match c with
      [] -> (s, reduced)
    | EqNames(v1, v2)::c' -> 
         if v1=v2 then reduce_it c' s reduced
         else 
           let s1 = Utilities.make_subst v1 v2 in
           let s2 = comp_nsubst s1 s in
           let c'' = app_subst_for_c s1 c' in
           let reduced' = app_subst_for_c s1 reduced in
             reduce_it c'' s2 reduced'
   | ac::c' -> 
             reduce_it c' s (ac::reduced) 
  in
    reduce_it c (Utilities.subst_empty) [];;

let rec get_bound c =
  let rec get_bound_it c bound rest =
    match c with
     [] -> (bound, rest)
   | (UpBound(x, nset))::c' ->
        (try
          let (nset1, bound') = Utilities.list_assoc2 x bound in
          let nset' = List.filter (fun x->List.mem x nset1) nset in
            get_bound_it c' ((x, nset')::bound') rest
        with
          Not_found -> get_bound_it c' ((x,nset)::bound) rest)
   | ac::c' -> get_bound_it c' bound (ac::rest)
  in
    let (bound, c') = get_bound_it c [] [] in
    let vars = nvars_in_c c' in
    let bounded = List.map fst bound in
    let unbounded = List.filter (fun x -> not(List.mem x bounded)) vars in
    let newbound = List.map (fun x -> (x, [])) unbounded in
      (newbound@bound, c')

let rec gfp bound c =
  let subset s1 s2 =
      List.for_all (fun x->List.mem x s2) s1
  in
  let app_subst s n =
     try
       List.assoc n s 
     with
       Not_found -> [n]
  in
  let rec gfp_one_step bound c flag =
    match c with
      [] -> (bound, flag)
    | SubNS(x, nsubst,y)::c' ->
        let (nset_y, bound') = try Utilities.list_assoc2 y bound with Not_found -> ([], bound) in
        let nset_x = try List.assoc x bound with Not_found -> [] in
        let nset_y' = List.filter (fun n -> subset (app_subst nsubst n) nset_x) nset_y in
          if nset_y=nset_y' then
            gfp_one_step bound c' flag
          else
            gfp_one_step ((y,nset_y')::bound') c' true
    | SubNS2(nsubst,x, y)::c' ->
        let (nset_y, bound') = try Utilities.list_assoc2 y bound with Not_found -> ([], bound) in
        let nset_x = try List.assoc x bound with Not_found -> [] in
        let nset_x' = List.flatten (List.map (app_subst nsubst) nset_x) in
        let nset_y' = List.filter (fun n -> List.mem n nset_x') nset_y in
          if nset_y=nset_y' then
            gfp_one_step bound c' flag
          else
            gfp_one_step ((y,nset_y')::bound') c' true
   | _ -> raise (Fatal "gfp")
  in
     let (bound', flag) = gfp_one_step bound c false in
       if flag then
           gfp bound' c
       else
           bound'

let rec expand_with_alias nl =
  match nl with
    [] -> []
  | n::nl' -> 
      let nl1 = List.map fst (List.filter (fun (x,y)->y=n) !alias_tab) in
        n::(nl1@(expand_with_alias nl'))

let is_var s = not(String.contains s '#')
let rec extract_order p =
  match p with
    Zero -> []
  | In(x,y,t,p1,_) ->
      let ord1 =
         match t with 
          ChanT(_, NConst nl) ->
            let nl' = List.filter is_var nl in
              List.map (fun n -> (n, y)) nl'
         | _ -> []
      in
      let ord2 = extract_order p1 in
         ord1@ord2
  | Out(_,_,p1,_) -> extract_order p1
  | Par(p1,p2) -> (extract_order p1)@(extract_order p2)
  | Rep(p1) -> extract_order p1
  | Nu(x, t, p1, _) ->
      let ord1 =
         match t with 
          ChanT(_, NConst nl) ->
            let nl' = List.filter is_var nl in
              List.map (fun n -> (n, x)) nl'
          | _ -> []
      in
      let ord2 = extract_order p1 in
         ord1@ord2
  | If(_,p1,p2,_) -> (extract_order p1)@(extract_order p2)
  | MatchL(_, ml, _) ->
      List.flatten (List.map (fun (_,p) -> (extract_order p)) ml)
  | Let(x,t,_, p1,_) ->
      let ord1 =
         match t with 
          ChanT(_, NConst nl) ->
            let nl' = List.filter is_var nl in
              List.map (fun n -> (n, x)) nl'
          | _ -> []
      in
      let ord2 = extract_order p1 in
         ord1@ord2
  | Special(p1) -> extract_order p1

let elaborate (p, stenv) =
  let _ = reset() in
 (*** let p' = Syntax.alphaconv p in ***)
 (*** let (p1, stenv) = Stype.elaborate p in ***)
  let tenv = Utilities.env_map (fun st -> new_otype st) stenv in
  let tenv' = Utilities.env2list tenv in
  let venv' = List.map (fun (x,t) -> (x, mkNS x t)) tenv' in
  let venv = Utilities.list2env venv' in
  let nset = nset_of_venv venv' in
  let (p1', c1) =  tinf p tenv venv nset in
  let c2 = merge_constraints (c1::(List.map (fun (x,t)-> mkC_Tboundempty(t)) tenv')) in
  let (s, c3) = reduce_eqN c2 in
  let p1'' = Syntax.map_for_t (app_nsubst_for_t s) p1' in
(***  let tenv'' = Utilities.env_map (app_nsubst_for_t s) tenv in ***)
  let (bound, c4) = get_bound c3 in
  let bound' = gfp bound c4 in
  let bound'' = List.map (fun (x, ns) -> (x, expand_with_alias ns)) bound' in
  let p2 = Syntax.map_for_t (subst_nconst_for_t bound'') p1'' in
(***  let tenv2 = Utilities.env_map (subst_nconst_for_t bound'') tenv'' in ***)
  let ord = extract_order p2 in
    (p2, ord)
